/*-
 *  Copyright (c) 2001  Peter Pentchev
 *  All rights reserved.
 * 
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xmu/StdSel.h>
#include <X11/Xmd.h>
#include <X11/extensions/xf86vmode.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

/* Some constants, shamelessly stolen from the xvidtune source */

/* Minimum extension version required */
#define MINMAJOR 0
#define MINMINOR 5

/* Mode flags -- ignore flags not in V_FLAG_MASK */
#define V_FLAG_MASK     0x1FF;
#define V_PHSYNC        0x001
#define V_NHSYNC        0x002
#define V_PVSYNC        0x004
#define V_NVSYNC        0x008
#define V_INTERLACE     0x010
#define V_DBLSCAN       0x020
#define V_CSYNC         0x040
#define V_PCSYNC        0x080
#define V_NCSYNC        0x100

/* Function prototypes */

int		 ShowMode(Display *, int);
int		 ParseModeInfo(XF86VidModeModeInfo *, int, char *[]);
int		 SetMode(Display *, int, XF86VidModeModeInfo *);

/* Flag representations in the xvidtune output */
struct vidflags {
	const char	*name;
	int		 flag;
} flags[] = {
	{"+hsync",	V_PHSYNC},
	{"-hsync",	V_NHSYNC},
	{"+vsync",	V_PVSYNC},
	{"-vsync",	V_NVSYNC},
	{"interlace",	V_INTERLACE},
	{"composite",	V_CSYNC},
	{"+csync",	V_PCSYNC},
	{"-csync",	V_NCSYNC},
	{"doublescan",	V_DBLSCAN},
	{NULL,		0}
};

static const char _cvs_id[] = "$Ringlet: c/x11/xvidsetmode/xvidsetmode.c,v 1.1 2004/07/14 16:56:56 roam Exp $";

int
ShowMode(Display *disp, int scrn)
{
	int dot_clock;
	XF86VidModeModeLine mode_line;
	
	if (!XF86VidModeGetModeLine(disp, scrn, &dot_clock, &mode_line)) {
		fprintf(stderr, "Could not get the current mode line\n");
		return 3;
	}
	printf("Mode: %dx%d\n", mode_line.hdisplay, mode_line.vdisplay);
	return (0);
}

int
ParseModeInfo(XF86VidModeModeInfo *dst, int argc, char *argv[])
{
	XF86VidModeModeInfo info;
	int i, j, fl;
#define	POPD(v)	(v) = strtol(argv[i++], NULL, 10)
	
	if (argc < 9) {
		fprintf(stderr, "Need at least 9 arguments\n");
		return (0);
	}
	memset(&info, 0, sizeof(info));

	info.dotclock = ceil(1000 * strtod(argv[1], NULL));
	i = 2;

	POPD(info.hdisplay);
	POPD(info.hsyncstart);
	POPD(info.hsyncend);
	POPD(info.htotal);

	POPD(info.vdisplay);
	POPD(info.vsyncstart);
	POPD(info.vsyncend);
	POPD(info.vtotal);

	/* Okay, the flags now... */
	fl = 0;
	for (; i < argc; i++) {
		for (j = 0; flags[j].name != NULL; j++)
			if (!strcmp(flags[j].name, argv[i]))
				break;
		if (flags[j].name == NULL) {
			fprintf(stderr, "Unknown mode flag '%s'\n", argv[i]);
			return (0);
		}
		fl |= flags[j].flag;
	}
	info.flags = fl;

	printf("Parsed a %.2f %dx%d (%d-%d=%d, %d-%d=%d, %04X) mode.\n",
	    (double)info.dotclock / 1000,
	    info.hdisplay, info.vdisplay,
	    info.hsyncstart, info.hsyncend, info.htotal,
	    info.vsyncstart, info.vsyncend, info.vtotal,
	    info.flags);
	memcpy(dst, &info, sizeof(*dst));
	return (1);
}

int
SetMode(Display *disp, int scrn, XF86VidModeModeInfo *info)
{
	int res;
	
	if (!XF86VidModeLockModeSwitch(disp, scrn, FALSE)) {
		printf("Unlocking mode switch failed\n");
		return (0);
	}
	XFlush(disp);

	res = XF86VidModeValidateModeLine(disp, scrn, info);
	if (res != 0) {
		printf("ValidateModeLine failed, result code %d\n", res);
		return (0);
	}

	printf("About to switch...\n");
	if (!XF86VidModeLockModeSwitch(disp, scrn, FALSE)) {
		printf("Unlocking mode switch failed\n");
		return (0);
	}
	XFlush(disp);
	if (!XF86VidModeSwitchToMode(disp, scrn, info)) {
		printf("SwitchToMode failed!\n");
		return (0);
	}
	XSync(disp, TRUE);
	return (1);
}

int
main(int argc, char *argv[])
{
	Widget top;
	XtAppContext app;
	Display *dpy;
	int scrn;
	int MajorVersion, MinorVersion;
	int EventBase, ErrorBase;
	XF86VidModeModeInfo modeinfo;

	if (argc < 2) {
		fprintf(stderr, "Usage:\txvidsetmode modeline\n"
		    "\txvidsetmode -show\n");
		return 1;
	}
	
	top = XtVaOpenApplication(&app, "Xvidsetmode", NULL, 0, &argc, argv,
	    NULL, applicationShellWidgetClass, XtNmappedWhenManaged,
	    False, NULL);
	dpy = XtDisplay(top);
	scrn = DefaultScreen(dpy);

	if (!XF86VidModeQueryVersion(dpy, &MajorVersion, &MinorVersion))
	{
		fprintf(stderr, "Unable to query video extension version\n");
		return 2;
	}
	if (!XF86VidModeQueryExtension(dpy, &EventBase, &ErrorBase))
	{
		fprintf(stderr,
		    "Unable to query video extension information\n");
		return 2;
	}
	printf("XFree86-VidMode %d.%d, eventbase %d, errorbase %d\n",
	    MajorVersion, MinorVersion, EventBase, ErrorBase);

	/* Fail if the extension version in the server is too old */
	if (MajorVersion < MINMAJOR ||
	    (MajorVersion == MINMAJOR && MinorVersion < MINMINOR))
	{
		fprintf(stderr,
		    "Xserver is running an old XFree86-VidModeExtension version"
		    " (%d.%d)\n", MajorVersion, MinorVersion);
		fprintf(stderr, "Minimum required version is %d.%d\n",
		    MINMAJOR, MINMINOR);
		exit(2);
	}

	if (!strcmp(argv[1], "-show"))
		return (ShowMode(dpy, scrn));
	if (!ParseModeInfo(&modeinfo, argc - 1, argv + 1))
		return (1);
	if (!SetMode(dpy, scrn, &modeinfo))
		return (1);

	printf("Done... it seems :)\n");
	return (0);
}
